#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author:  C. Frescolino, D. Gresch
# File:    __init__.py

"""
This is a helper for saving and loading data from files, with a given encoding and decoding function. It can automatically detect the appropriate serializer based on the file ending, and saves in an atomic way by first saving to a temporary file and then moving it.

To define an encoding, an instance of :class:`.SerializerDispatch` is created, which has methods for saving and loading.

A default encoding which can handle common numpy and built-in types is also given.
"""

from ._version import __version__

from . import encoding
from ._iohelper import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author:  Dominik Gresch <greschd@gmx.ch>
# Date:    08.03.2016 10:21:06 CET
# File:    _save_load.py

"""Free functions for saving and loading objects with a given encoding."""

import os
import json
import pickle
import tempfile
from collections import namedtuple

import msgpack

from fsc.export import export

SerializerSpecs = namedtuple('SerializerSpecs', ['binary', 'encode_kwargs', 'decode_kwargs'])

@export
class SerializerDispatch:
    """
    Defines functions for saving and loading for a given encoding.

    :param encoding: Object with ``encode`` and ``decode`` members, to be passed as the ``default`` and ``object_hook`` parameters, respectively (for JSON and msgpack, see e.g. :py:func:`json.dump` and :py:func:`json.load`).

    :param exclude: List of excluded serializers.
    :type exclude: list


    Basic usage:

    .. code :: python

        import fsc.iohelper

        IO_HANDLER = fsc.iohelper.SerializerDispatch(fsc.iohelper.encoding.default)
        IO_HANDLER.save([1, 2, 3], 'filename.json')
        x = IO_HANDLER.load('filename.json')
    """
    def __init__(self, encoding, exclude=[]):
        self.ext_mapping = {
            k.lower(): v for k, v in [
                ('p', pickle),
                ('pickle', pickle),
                ('msgpack', msgpack),
                ('json', json)
            ]
            if v not in exclude
        }
        self.serializer_specs = dict(
            [
                (json, SerializerSpecs(
                    binary=False,
                    encode_kwargs=dict(default=encoding.encode),
                    decode_kwargs=dict(object_hook=encoding.decode)
                )),
                (msgpack, SerializerSpecs(
                    binary=True,
                    encode_kwargs=dict(default=encoding.encode),
                    decode_kwargs=dict(object_hook=encoding.decode, encoding='utf-8')
                )),
                (pickle, SerializerSpecs(
                    binary=True,
                    encode_kwargs=dict(protocol=4),
                    decode_kwargs=dict()
                ))
            ]
        )
        for excluded_serializer in exclude:
            self.serializer_specs.pop(excluded_serializer, None)

    def _get_serializer(self, file_path, use_default=False):
        """Tries to determine the correct serializer from the file extension. If none can be determined, falls back to the default (:py:mod:`json`) if ``use_default`` is true, otherwise it throws an error."""
        _, file_ext = os.path.splitext(file_path)
        try:
            return self.ext_mapping[file_ext.lower().lstrip('.')]
        except KeyError:
            if use_default:
                return json
            else:
                raise ValueError("Could not guess serializer from file ending '{}'".format(file_ext))

    def save(self, obj, file_path, serializer='auto'):
        """Saves an object to the file given in ``file_path``. The saving is made atomic (on systems where :py:func:`os.replace` is atomic) by first creating a temporary file and then moving to the ``file_path``.

        :param obj:         Object to be saved.

        :param file_path:   Path to the file.
        :type file_path:    str

        :param serializer:  The serializer to be used. Valid options are :py:mod:`msgpack` :py:mod:`json` and :py:mod:`pickle`. By default, the serializer is determined from the file extension. If this does not work, :py:mod:`json` is used to avoid data loss.
        :type serializer:   module
        """
        if serializer == 'auto':
            serializer = self._get_serializer(file_path, use_default=True)
        specs = self.serializer_specs[serializer]
        dirname = os.path.dirname(os.path.abspath(file_path))
        if not os.path.isdir(dirname):
            raise ValueError('Directory {} does not exist'.format(dirname))
        try:
            with tempfile.NamedTemporaryFile(
                dir=dirname,
                delete=False,
                mode='wb' if specs.binary else 'w'
            ) as f:
                tmp_path = f.name
                serializer.dump(obj, f, **specs.encode_kwargs)
            # closing necessary on Windows
            os.replace(tmp_path, file_path)
        except Exception as e:
            os.remove(tmp_path)
            raise e

    def load(self, file_path, serializer='auto'):
        """Loads the object that was saved to ``file_path``.

        :param file_path:   Path to the file.
        :type file_path:    str

        :param serializer:  The serializer which should be used to load the result. By default, is deduced from the file extension. If no serializer is given and it cannot be deduced from the file ending, a :py:class:`ValueError` is raised, to avoid loading corrupted data.
        :type serializer:   module
        """
        if serializer == 'auto':
            serializer = self._get_serializer(file_path)
        specs = self.serializer_specs[serializer]
        with open(file_path, 'rb' if specs.binary else 'r') as f:
            return serializer.load(f, **specs.decode_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author:  Dominik Gresch <greschd@gmx.ch>
# Date:    21.09.2016 11:37:00 CEST
# File:    __init__.py

"""
This module contains the encodings that are defined within iohelper.
"""

from . import default



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author:  Dominik Gresch <greschd@gmx.ch>
# Date:    15.04.2016 10:21:56 CEST
# File:    default.py

"""
This is an example implementation of an encoding, to be used as input for :class:`.SerializerDispatch`.
"""

import numbers
import contextlib
from functools import singledispatch
from collections.abc import Iterable

import numpy as np
from fsc.export import export

@export
@singledispatch
def encode(obj):
    """
    Encodes numpy and built-in types into JSON / msgpack - compatible types.
    """
    raise TypeError('cannot JSONify {} object {}'.format(type(obj), obj))

@encode.register(np.bool_)
def _(obj):
    return bool(obj)

@encode.register(numbers.Real)
def _(obj):
    return float(obj)

@encode.register(numbers.Complex)
def _(obj):
    return dict(__complex__=True, real=encode(obj.real), imag=encode(obj.imag))

@encode.register(Iterable)
def _(obj):
    return list(obj)

#-----------------------------------------------------------------------#

@export
@singledispatch
def decode(obj):
    """
    Decodes JSON / msgpack objects into the corresponding built-in types.
    """
    return obj

def decode_complex(obj):
    return complex(obj['real'], obj['imag'])

@decode.register(dict)
def _(obj):
    with contextlib.suppress(AttributeError):
        obj = {k.decode('utf-8'): v for k, v in obj.items()}
    special_markers = [key for key in obj.keys() if key.startswith('__')]
    if len(special_markers) == 1:
        name = special_markers[0].strip('__')
        return globals()['decode_' + name](obj)
    else:
        return obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author:  Dominik Gresch <greschd@gmx.ch>
# Date:    30.03.2016 12:21:07 CEST
# File:    _version.py

import os

with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'version.txt'), 'r') as f:
    __version__ = f.read().strip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author:  Dominik Gresch <greschd@gmx.ch>
# Date:    30.03.2016 00:29:06 CEST
# File:    setup.py

import sys

from setuptools import setup

pkgname = 'iohelper'
pkgname_qualified = 'fsc.' + pkgname

with open('doc/description.txt', 'r') as f:
    description = f.read()
try:
    with open('doc/README', 'r') as f:
        readme = f.read()
except IOError:
    readme = description

with open('version.txt', 'r') as f:
    version = f.read().strip()
    
requirements = ['msgpack-python', 'fsc.export', 'numpy']
if sys.version_info < (3,):
    raise ValueError('only Python 3.x and higher are supported')

setup(
    name=pkgname_qualified,
    version=version,
    packages=[
        pkgname_qualified,
        pkgname_qualified + '.encoding'
    ],
    url='http://frescolinogroup.github.io/frescolino/pyiohelper/' + '.'.join(version.split('.')[:2]),
    include_package_data=True,
    author='C. Frescolino',
    author_email='frescolino@lists.phys.ethz.ch',
    description=description,
    install_requires=requirements,
    long_description=readme,
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities'
    ],
    license='Apache',
)
